#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

void benc(char *dst, char *s);
void encbasic(char *d, char *u, char *p);
void die (char *s);
void open_conn(int s, char *hostname, unsigned int port);

enum {
	kHOST, kMID, kPASS
};

enum {
	kPort = 80
};

#define kScriptPath "/mis-bin/ms3/nic/dyndns?command=Update+Host&domain=%s&act=act&wildcard=on&do=mod&agree=agree"
#define kScriptHost "members.ml.org"

static const char BaseTable[] =
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		"abcdefghijklmnopqrstuvwxyz"
		"0123456789+/";

void
benc (char *dst, char *s)
{
	int n = strlen(s), n3byt = n / 3, k = n3byt * 3, nrest = n % 3, i = 0, dstlen = 0;

	while (i < k) {
		dst[dstlen++] = BaseTable[(char)(( s[i]   & 0xFC) >> 2)];
		dst[dstlen++] = BaseTable[(char)(((s[i]   & 0x03) << 4) | ((s[i+1] & 0xF0) >> 4))];
		dst[dstlen++] = BaseTable[(char)(((s[i+1] & 0x0F) << 2) | ((s[i+2] & 0xC0) >> 6))];
		dst[dstlen++] = BaseTable[(char)(  s[i+2] & 0x3F)];
		i += 3;
	}

	if (nrest == 2) {
		dst[dstlen++] = BaseTable[(char)(( s[k] & 0xFC)   >> 2)];
		dst[dstlen++] = BaseTable[(char)(((s[k] & 0x03)   << 4) | ((s[k+1] & 0xF0) >> 4))];
		dst[dstlen++] = BaseTable[(char)(( s[k+1] & 0x0F) << 2)];
	} else if (nrest == 1) {
		dst[dstlen++] = BaseTable[(char)((s[k] & 0xFC) >> 2)];
		dst[dstlen++] = BaseTable[(char)((s[k] & 0x03) << 4)];
	}

	while (dstlen % 3)
		dst[dstlen++] = '=';

	dst[dstlen] = '\0';
}

void encbasic(char *d, char *u, char *p)
{
	char _buf[4096];

	strcpy(_buf, u);
	strcat(_buf, ":");
	strcat(_buf, p);
	benc(d, _buf);
}

void
open_conn (int s, char *hostname, unsigned int port)
{
	struct sockaddr_in sin;
	struct hostent *he;

	memset(&sin, 0, sizeof sin);

	sin.sin_family	= AF_INET;
	sin.sin_port	= htons(port);

	if (!(he = gethostbyname(hostname)))
		die("unable to resolve address");

	memcpy(&sin.sin_addr, he->h_addr, he->h_length);

	if (connect(s, (struct sockaddr*)&sin, sizeof(sin)) < 0)
		die("connect");
}

void
die (char *err)
{
	perror(err);
	exit(-1);
}

#define _DEFAULT_FILE "/boot/.dyn.cfg"

int
main (int argc, char *const *argv)
{
	FILE *fp;
	char host[128], mid[128], pass[128];
	char buf[4096], scriptpath[4096];
	char cur = 0, varpos = 0;
	int i = 0, len, s;

	memset(&buf, 0, sizeof buf);
	memset(&host, 0, sizeof host);
	memset(&mid, 0, sizeof mid);
	memset(&pass, 0, sizeof pass);
	
	if (!(fp = fopen(argv[1] ? argv[1] : _DEFAULT_FILE, "r")))
		die(argv[1]?argv[1]:_DEFAULT_FILE);

	fread(buf, sizeof buf - 1, sizeof(char), fp);

	fclose(fp);

	while (cur < 3) {
		switch (cur) {
			case kHOST:
				for (;;) {
					if (buf[i] == '\n') {
						cur++;
						i++;
						varpos = 0;
						break;
					}
					if (!buf[i] || buf[i] == ' ' || buf[i] == '\t' || varpos > 127)
						die("reading hostname from file");
					host[varpos++] = buf[i++];
				}
				break;
			case kMID:
				for (;;) {
					if (buf[i] == '\n') {
						cur++;
						i++;
						varpos = 0;
						break;
					}
					if (!buf[i] || buf[i] == ' ' || buf[i] == '\t' || varpos > 127)
						die("reading mid from file");
					mid[varpos++] = buf[i++];
				}
				break;
			case kPASS:
				for(;;) {
					if (buf[i] == '\n' || !buf[i]
					|| buf[i] == ' ' || buf[i] == '\t' || varpos > 127) {
						cur++;
						break;
					}
					pass[varpos++] = buf[i++];
				}
				break;
		}
	}

	encbasic(pass, mid, pass);

	if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		die("socket");

	open_conn(s, kScriptHost, kPort);

	sprintf(scriptpath, kScriptPath, host);

	len = sprintf(buf,	"GET %s HTTP/1.0\r\n"
				"Pragma: no-cache\r\n"
				"User-Agent: dyn\r\n"
				"Authorization: Basic %s\r\n"
				"\r\n",
			scriptpath, pass);

	if ((i = send(s, buf, len, 0)) != len)
		die("send");

	while ((i = recv(s, buf, sizeof buf, 0)) > 0)
		write(1, buf, i);

	close(s);

	return 0;
}
